unit timeedit;

interface

uses
  Windows, Classes, StdCtrls, ExtCtrls, Controls, Messages, SysUtils,
  Forms, Graphics, Menus, Buttons, Spin;

type
  TTimeEdit = class(TCustomEdit)
  private
    FMinValueHour: LongInt;
    FMaxValueHour: LongInt;
    FMinValueMin: LongInt;
    FMaxValueMin: LongInt;
    FIncHour: LongInt;
    FIncMin: LongInt;
    FButHour: TSpinButton;
    FButMin: TSpinButton;
    FEditorEnabled: Boolean;
    function GetMinHeight: Integer;
    function GetValueHour: LongInt;
    function GetValueMin: LongInt;
    function CheckValueHour (NewValue: LongInt): LongInt;
    function CheckValueMin (NewValue: LongInt): LongInt;
    procedure SetValueHour (NewValue: LongInt);
    procedure SetValueMin (NewValue: LongInt);
    procedure SetEditRect;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMExit);   message CM_EXIT;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
  protected
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure UpHourClick (Sender: TObject); virtual;
    procedure DownHourClick (Sender: TObject); virtual;
    procedure UpMinClick (Sender: TObject); virtual;
    procedure DownMinClick (Sender: TObject); virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Button: TSpinButton read FButHour;
  published
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property EditorEnabled: Boolean read FEditorEnabled write FEditorEnabled default True;
    property Enabled;
    property Font;
    property IncrementHour: LongInt read FIncHour write FIncHour default 1;
    property IncrementMin: LongInt read FIncMin write FIncMin default 1;
    property MaxLength;
    property MaxValueHour: LongInt read FMaxValueHour write FMaxValueHour default 23;
    property MinValueHour: LongInt read FMinValueHour write FMinValueHour default 0;
    property MaxValueMin: LongInt read FMaxValueMin write FMaxValueMin default 59;
    property MinValueMin: LongInt read FMinValueMin write FMinValueMin default 0;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property ValueHour: LongInt read GetValueHour write SetValueHour;
    property ValueMin: LongInt read GetValueMin write SetValueMin;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

procedure Register;

implementation

constructor TTimeEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButHour := TSpinButton.Create (Self);
  FButHour.Width := 15;
  FButHour.Height := 17;
  FButHour.Visible := True;
  FButHour.Parent := Self;
  FButHour.FocusControl := Self;
  FButHour.OnUpClick := UpHourClick;
  FButHour.OnDownClick := DownHourClick;

  FButMin := TSpinButton.Create (Self);
  FButMin.Width := 15;
  FButMin.Height := 17;
  FButMin.Visible := True;
  FButMin.Parent := Self;
  FButMin.FocusControl := Self;
  FButMin.OnUpClick := UpMinClick;
  FButMin.OnDownClick := DownMinClick;

  Text := '00:00';
  ControlStyle := ControlStyle - [csSetCaption];
  FIncHour := 1;
  FIncMin := 1;
  FMaxValueHour:=23;
  FMaxValueMin:=59;
  FMinValueHour:=0;
  FMinValueMin:=0;
  FEditorEnabled := false;
end;

destructor TTimeEdit.Destroy;
begin
  FButHour := nil;
  FButMin := nil;
  inherited Destroy;
end;

procedure TTimeEdit.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
end;

procedure TTimeEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Key = VK_UP then UpHourClick (Self)
  else if Key = VK_DOWN then DownHourClick (Self);
  inherited KeyDown(Key, Shift);
end;

procedure TTimeEdit.KeyPress(var Key: Char);
begin
  if not IsValidChar(Key) then
  begin
    Key := #0;
    MessageBeep(0)
  end;
  if Key <> #0 then inherited KeyPress(Key);
end;

function TTimeEdit.IsValidChar(Key: Char): Boolean;
begin
  Result := (Key in [DecimalSeparator, '+', '-', '0'..'9']) or
    ((Key < #32) and (Key <> Chr(VK_RETURN)));
  if not FEditorEnabled and Result and ((Key >= #32) or
      (Key = Char(VK_BACK)) or (Key = Char(VK_DELETE))) then
    Result := False;
end;

procedure TTimeEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
{  Params.Style := Params.Style and not WS_BORDER;  }
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

procedure TTimeEdit.CreateWnd;
begin
  inherited CreateWnd;
  SetEditRect;
end;

procedure TTimeEdit.SetEditRect;
var
  Loc: TRect;
begin
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  Loc.Bottom := ClientHeight + 1;  {+1 is workaround for windows paint bug}
  Loc.Right := ClientWidth - FButHour.Width - FButMin.Width - 2;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));  {debug}
end;

procedure TTimeEdit.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  MinHeight := GetMinHeight;
    { text edit bug: if size to less than minheight, then edit ctrl does
      not display the text }
  if Height < MinHeight then
    Height := MinHeight
  else if FButHour <> nil then
  begin
    if NewStyleControls and Ctl3D then
    begin
      FButHour.SetBounds(Width - FButHour.Width - FButMin.Width - 5, 0, FButHour.Width, Height - 5);
      FButMin.SetBounds(Width - FButMin.Width - 5, 0, FButMin.Width, Height - 5);
    end
    else begin
      FButHour.SetBounds (Width - FButHour.Width - FButMin.Width, 1, FButHour.Width, Height - 3);
      FButMin.SetBounds (Width - FButMin.Width, 1, FButHour.Width, Height - 3);
    end;
    SetEditRect;
  end;
end;

function TTimeEdit.GetMinHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  Result := Metrics.tmHeight + I div 4 + GetSystemMetrics(SM_CYBORDER) * 4 + 2;
end;

procedure TTimeEdit.WMPaste(var Message: TWMPaste);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TTimeEdit.WMCut(var Message: TWMPaste);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TTimeEdit.CMExit(var Message: TCMExit);
begin
  inherited;
  if CheckValueHour (ValueHour) <> ValueHour then
    SetValueHour (ValueHour);
end;

procedure TTimeEdit.UpHourClick (Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else ValueHour := ValueHour + FIncHour;
end;

procedure TTimeEdit.DownHourClick (Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else ValueHour := ValueHour - FIncHour;
end;

procedure TTimeEdit.UpMinClick (Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else ValueMin := ValueMin + FIncMin;
end;

procedure TTimeEdit.DownMinClick (Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else ValueMin := ValueMin - FIncMin;
end;

function TTimeEdit.GetValueHour: LongInt;
begin
  try
    Result := StrToInt(copy(Text,1,2));
  except
    Result := FMinValueHour;
  end;
end;

function TTimeEdit.GetValueMin: LongInt;
begin
  try
    Result := StrToInt (copy(Text,4,2));
  except
    Result := FMinValueMin;
  end;
end;

procedure TTimeEdit.SetValueHour (NewValue: LongInt);
var LHour,Lmin:string;
begin
  if CheckValueHour(NewValue)<10 then LHour:='0' else LHour:='';
  if ValueMin<10 then LMin:='0' else LMin:='';

  Text := LHour+IntToStr(CheckValueHour(NewValue))+':'+LMin+IntToStr(ValueMin);
end;

procedure TTimeEdit.SetValueMin (NewValue: LongInt);
var LHour,Lmin:string;
begin
  if ValueHour<10 then LHour:='0' else LHour:='';
  if CheckValueMin(NewValue)<10 then LMin:='0' else LMin:='';

  Text := LHour+IntToStr(ValueHour)+':'+LMin+IntToStr(CheckValueMin(NewValue));
end;

function TTimeEdit.CheckValueHour (NewValue: LongInt): LongInt;
begin
  Result := NewValue;
  if (FMaxValueHour <> FMinValueHour) then
  begin
    if NewValue < FMinValueHour then
      Result := NewValue+FMaxValueHour+1
    else if NewValue > FMaxValueHour then
      Result := NewValue-FMaxValueHour-1;
  end;
end;

function TTimeEdit.CheckValueMin (NewValue: LongInt): LongInt;
begin
  Result := NewValue;
  if (FMaxValueMin <> FMinValueMin) then
  begin
    if NewValue < FMinValueMin then
      Result := NewValue+FMaxValueMin+1
    else if NewValue > FMaxValueMin then
      Result := NewValue-FMaxValueMin-1;
  end;
end;

procedure TTimeEdit.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect and not (csLButtonDown in ControlState) then
    SelectAll;
  inherited;
end;

procedure Register;
begin
 RegisterComponents('Samples', [TTimeEdit]);
end;

end.

